/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Session module.
 * 
 * See session.h and session-api.h for more information on the functions and types involved.
 */

#include <pthread.h>
#include <errno.h>
#include <assert.h>
#include <time.h>
#include <string.h>
#include <stdio.h>

#include "waaad-internal.h"

/* More details about the session implementation:
 *
 *  The sess_id_t, representing a session id, are stored as _sess_id_t objects in a hash table.
 * These objects contains generic information about the session: sid string, creation time, ...
 * They keep a refcount. when the refcount becomes 0, the object is freed.
 *
 *  The sess_reg_t represent clients for the module.
 * The data is stored in a list which root is this _sess_reg_t object.
 * Each element of the list contains a pointer to a _sess_id_t, a pointer to user's data, and
 * a reference to a cleanup handler.
 */

/* Types and globals definitions */

/* The following two integers are used to generate values which are eternaly unique */
static uint32_t   	g_sid_h;	/* initialized to the current time at module initialization */
static uint32_t   	g_sid_l;	/* incremented each time a session id is created. */
static pthread_mutex_t 	g_sid_lock = PTHREAD_MUTEX_INITIALIZER;

/************ sess_id_t ***********/

/* Size of the hash table to hold the sess_id_t objects (pow of 2. ex: 6 => 2^6 = 64). must be between 0 and 31. */
#define SESS_HASH_SIZE	6

/* How objects are chained. Each sub-list of the hash table is sorted by hash value then strcmp'd. */
typedef struct _sli_ {
	struct _sli_  * next;
	struct _sli_  * prev;
} _sli_t;

#define _SLI( _sli_ ) ((_sli_t *)(_sli_))

#define SLI_init( _sli ) {		\
	_SLI(_sli)->next = _SLI(_sli);	\
	_SLI(_sli)->prev = _SLI(_sli);	\
}

#define SLI_isempty( _sli ) ( _SLI(_sli)->next == _SLI(_sli) )

#define SLI_remove( _sli ) {				\
	_SLI(_sli)->next->prev = _SLI(_sli)->prev;	\
	_SLI(_sli)->prev->next = _SLI(_sli)->next;	\
	_SLI(_sli)->next = _SLI(_sli);			\
	_SLI(_sli)->prev = _SLI(_sli);			\
}

#define SLI_addafter( _ref, _sli ) {		\
	assert(SLI_isempty( _sli ));		\
	_SLI(_sli)->next = _SLI(_ref)->next;	\
	_SLI(_sli)->prev = _SLI(_ref);		\
	_SLI(_ref)->next->prev = _SLI(_sli);	\
	_SLI(_ref)->next = _SLI(_sli);		\
}
	
	
/* The sess_id_t internal definition */
typedef struct {
	_sli_t	 	 chain;	/* Chaining information */
	uint32_t	 hash;	/* The value of the hash of the sid string */
	int		 eyec;	/* An eye catcher to ensure the object is valid */
	char 		*sid;	/* The session id string, \0 terminated. */
	int		 rc;	/* The refcount */
	struct timespec	 ts;	/* Time of the object creation */
} _sess_id_t;

/* The eyec value */
#define _SI_EYEC 0x53551D

/* some useful cast macros */
#define _SI( _si ) ((_sess_id_t *)(_si))

/* Macro to check an element is valid */
#define VALIDATE_SI( _si ) ( ((_si) != NULL) && ( _SI(_si)->eyec == _SI_EYEC) )

/* The hash table */
static struct {
	_sli_t		sentinel;	/* sentinel element for this sublist */
	pthread_mutex_t lock;		/* the mutex for this sublist */
	int		count;		/* Number of items in this list */
} sess_hash [ 1 << SESS_HASH_SIZE ] ;
#define H_MASK( __hash ) ((__hash) & (( 1 << SESS_HASH_SIZE ) - 1))
#define H_LIST( _hash ) (&(sess_hash[H_MASK(_hash)].sentinel))
#define H_LOCK( _hash ) (&(sess_hash[H_MASK(_hash)].lock    ))
#define H_COUNT(_hash ) (  sess_hash[H_MASK(_hash)].count    )


/************ sess_reg_t and data ***********/

/* A list element */
typedef struct {
	_sli_t		  list;		 /* Elements are ordered in the same order as the hash table: by hash and string */
	_sess_id_t	 *si;		 /* pointer to the session element this list element refers */
	void 		(*clncb)(void *);/* The cleanup callback to call on data when the element is destroyed */
	void		 *data;		 /* The data that is registered for this client and session */
} _sd_data_t;

#define _SD( _sd_ ) ( (_sd_data_t *)( _sd_) )

/* Now the sess_reg_t internal definition */
typedef struct {
	_sli_t		list;	 /* List of the registered clients with no particular order; only to ensure proper cleanup in the end. */
	int		eyec;	 /* An eye catcher to ensure the object is valid */
	pthread_mutex_t lock;	 /* A lock to protect the list of data */
	_sli_t		sentinel;/* Sentinel for the list of _sd_data_t of this client */
} _sess_reg_t;

/* The sentinel for the list of registered handlers */
static _sli_t _srt_sentinel;
static pthread_mutex_t _str_lock; /* lock to protect the list of registered clients */

/* The eyec value */
#define _SR_EYEC 0x5355DA1A

/* useful cast macro */
#define _SR( _sr ) ((_sess_reg_t *)(_sr))

/* Macro to check an element is valid */
#define VALIDATE_SR( _sr ) ( ((_sr) != NULL) && ( _SR(_sr)->eyec == _SR_EYEC) )

	
/********************************************************************************************************/
/* Hash function -- credits to Austin Appleby, thank you ^^ */
/* See http://murmurhash.googlepages.com for more information on this function */

/* Our sid is always aligned properly, so we use the simple MurmurHash2 function */
#define _HASH_MIX(h,k,m) { k *= m; k ^= k >> r; k *= m; h *= m; h ^= k; }
static uint32_t _hash ( char * sid, size_t len )
{
	uint32_t hash = len;
	char * data = sid;
	
	const unsigned int m = 0x5bd1e995;
	const int r = 24;
	
	/* Check the alignment is really correct -- otherwise we have to switch to the other version */
	assert(((long)sid & 3) == 0);

	while(len >= 4)
	{
		/* Mix 4 bytes at a time into the hash */
		uint32_t k = *(uint32_t *)data;	/* We don't care about the byte order */
		
		_HASH_MIX(hash, k, m);

		data += 4;
		len -= 4;
	}
	
	/* Handle the last few bytes of the input */
	switch(len) {
		case 3: hash ^= data[2] << 16;
		case 2: hash ^= data[1] << 8;
		case 1: hash ^= data[0];
	        	hash *= m;
	}

	/* Do a few final mixes of the hash to ensure the last few
	   bytes are well-incorporated. */
	hash ^= hash >> 13;
	hash *= m;
	hash ^= hash >> 15;

	return hash;
} 
/********************************************************************************************************/

/* Function to compare a (hash, sid, len) with a sess_id_element. */
static __inline__ int cmp_si(uint32_t hash1, char * sid1, size_t sid1len,  _sess_id_t * si2)
{
	if (hash1 < si2->hash)
		return -1;
	if (hash1 > si2->hash)
		return 1;
	return strncmp(sid1, si2->sid, sid1len); /* si2->sid is NULL terminated */
}

/* Function to find a location for a (hash, sid) in a _sess_id_t list.
 * return 0 if the element is found, -1 if the previous element has been returned (in *loc)
 * The sublist lock must be held */
static __inline__ int find_si(uint32_t hash, char * sid, size_t sidlen, _sli_t * sentinel, _sli_t ** loc )
{
	int cmp = -1;
	for ((*loc) = sentinel; (*loc)->next != sentinel; (*loc) = (*loc)->next) {
		cmp = cmp_si(hash, sid, sidlen, _SI((*loc)->next));
		if (!cmp)
			(*loc) = (*loc)->next; /* we found the sid in the list */
		if (cmp <= 0)	/* The next element in the list is bigger than our element, so stop here */
			break;
	}
	return cmp;
}
	
/* Function to find a location for a _sess_id_t in a _sd_data_t list.
 * return 0 if the element is found, -1 if the previous element has been returned (in *loc)
 * The list lock must be held */
static __inline__ int find_sd(_sess_id_t * sid, _sli_t * sentinel, _sli_t ** loc )
{
	int cmp = -1;
	size_t len = strlen(sid->sid);
	for ((*loc) = sentinel; (*loc)->next != sentinel; (*loc) = (*loc)->next) {
		cmp = cmp_si(sid->hash, sid->sid, len, _SD((*loc)->next)->si);
		if (!cmp)
			(*loc) = (*loc)->next; /* we found the sid in the list */
		if (cmp <= 0)	/* The next element in the list is bigger than our element, so stop here */
			break;
	}
	return cmp;
}
	
	

/********************************************************************************************************/

/* Remove a piece of data linked from a _sess_reg_t. The _sess_reg_t lock must be held */
static int _destroy_sr_data(_sd_data_t * sd, void ** data)
{
	int ret = 0;

	/* Call the destroy function for this element */
	if (data)
		*data = sd->data;
	else
		if (sd->clncb)
			sd->clncb(sd->data);

	/* Unlink the session */
	ret = sess_unlink((sess_id_t **)&sd->si);
	if (ret != 0) {
		TRACE_DEBUG(INFO, "sess_unlink failed: %s", strerror(ret));
		return ret;
	}

	/* Unchain this element */
	SLI_remove( sd );

	/* And free it */
	free(sd);
	
	return 0;
}

/* Remove any data linked from a _sess_reg_t. The _sess_reg_t lock must be held */
static int _destroy_sr_alldata(_sli_t * sentinel)
{
	int ret = 0;
	while (!SLI_isempty( sentinel )) {
		ret = _destroy_sr_data(_SD(sentinel->next), NULL);
		if (ret != 0) {
			TRACE_DEBUG(INFO, "_destroy_sr_data failed: %s", strerror(ret));
			return ret;
		}
	}
	return 0;
}

/* Remove a _sess_reg_t and any data linked from it. The _str_lock lock must be held */	
static int _destroy_sr(_sess_reg_t * item)
{
	int ret = 0;
	
	/* Lock the object */
	ret = pthread_mutex_lock( &item->lock );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_lock failed: %s", strerror(ret));
		return ret;
	}
	
	/* Invalidate */
	item->eyec = 0xdead;
	
	/* Destroy any data contained in this object */
	ret = _destroy_sr_alldata(&item->sentinel);
	if (ret != 0) {
		TRACE_DEBUG(INFO, "_destroy_sr_alldata failed: %s", strerror(ret));
		(void)pthread_mutex_unlock( &item->lock );
		return ret;
	}
	
	/* Unlink */
	SLI_remove(item);
	
	/* Destroy the mutex */
	ret = pthread_mutex_unlock( &item->lock );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
		return ret;
	}
	ret = pthread_mutex_destroy( &item->lock );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_destroy failed: %s", strerror(ret));
		return ret;
	}
	
	/* Free the object */
	free(item);
	
	return 0;
}
	

/********************************************************************************************************/

/* Initialize the module */
int sess_init ( void )
{
	int i;
	
	TRACE_ENTRY( "" );
	
	/* Initialize the hash table */
	for (i = 0; i < sizeof(sess_hash) / sizeof(sess_hash[0]); i++) {
		int ret = pthread_mutex_init(&sess_hash[i].lock, NULL);
		if (ret != 0) {
			TRACE_DEBUG(INFO, "pthread_mutex_init failed: %s", strerror(ret));
			return ret;
		}
		SLI_init( &sess_hash[i].sentinel );
		sess_hash[i].count = 0;
	}
	
	/* Initialize the sentinel for registered clients */
	SLI_init( &_srt_sentinel );
	/* and its lock */
	i = pthread_mutex_init(&_str_lock, NULL);
	if (i != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_init failed: %s", strerror(i));
		return i;
	}
	
	/* Initialize the global counters */
	g_sid_h = (uint32_t) time(NULL);
	g_sid_l = 0;
	
	return 0;
}

/* Find a session object corresponding to a given session-id / or create it */
int sess_fromsid ( char * sid, size_t len, sess_id_t ** session, int * new )
{
	int      ret = 0;
	_sli_t	*li  = NULL;
	uint32_t hash;
	
	TRACE_ENTRY( "%p %d %p", sid, len, session );
	
	/* Validate the parameters */
	if ( (sid == NULL) || (len == 0) || (session == NULL) ) {
		TRACE_DEBUG(INFO, "Invalid parameters");
		return EINVAL;
	}
	
	/* Compute the hash value of this string */
	hash = _hash(sid, len);
	
	/* Lock the hash table line */
	ret = pthread_mutex_lock( H_LOCK(hash) );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_lock failed: %s", strerror(ret));
		return ret;
	}
	
	/* Search in the table */
	ret = find_si( hash, sid, len, H_LIST(hash), &li);
	if (ret == 0) {
		/* The element is found in the table */
		if (new)
			*new = 0;
		
		/* increment its refcount, as if sess_link had been called */
		_SI(li)->rc++;
		
		/* Save the value that must be returned */
		*session = (sess_id_t *)li; 
	} else {
		_sess_id_t * newsi;
		
		/* The element did not exist already, li points to the previous element */
		/* Create a new element -- how to do this outside the mutex section, and efficiently? */
		newsi = malloc(sizeof(_sess_id_t));
		if (newsi == NULL) {
			log_error("Memory allocation failed: %s\n", strerror(errno));
			TRACE_DEBUG(INFO, "malloc failed");
			ret = ENOMEM;
			goto end;
		}
		
		/* Initialize the element content */
		memset(newsi, 0, sizeof(_sess_id_t));
		SLI_init( &newsi->chain );
		newsi->hash = hash;
		newsi->eyec = _SI_EYEC;
		newsi->sid  = malloc(len + 1);
		if (newsi->sid == NULL) {
			log_error("Memory allocation failed: %s\n", strerror(errno));
			TRACE_DEBUG(INFO, "malloc failed");
			ret = ENOMEM;
			free(newsi);
			goto end;
		}
		memcpy(newsi->sid, sid, len);
		newsi->sid[len] = '\0';
		newsi->rc = 1;
		ret = clock_gettime(CLOCK_REALTIME, &newsi->ts);
		if (ret != 0) {
			TRACE_DEBUG(INFO, "clock_gettime failed: %s", strerror(ret));
			free(newsi->sid);
			free(newsi);
			goto end;
		}
			
		/* Store this element in the list */
		SLI_addafter(li, newsi);
		
		/* Increment the counter */
		H_COUNT(hash) += 1;
		
		/* Save the reference to it */
		*session = (sess_id_t *)newsi; 
		
		if (new)
			*new = 1;
	}
end:	
	/* Finish, unlock the line */
	{
		int __ret = pthread_mutex_unlock( H_LOCK(hash) );
		if (__ret != 0) {
			TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(__ret));
			return __ret;
		}
	}
	
	return ret;
}

/* Create a new session (identifier and object). */
int sess_new ( sess_id_t ** session, sess_flags_t flags, char * opt )
{
	int ret = 0;
	int new = 0;
	char * str = NULL;
	size_t strsz = 0;
	int mustfree = 0;
	
	TRACE_ENTRY( "%p %d %p", session, flags, opt );
	
	/* Check the parameters */
	if (session == NULL) {
		TRACE_DEBUG(INFO, "Invalid parameter");
		return EINVAL;
	}
	
	/* Ok, first create the identifier following the flags */
	switch (flags) {
		case SESSION_NEW_DEFAULT: 	/* "<diameterId>;<high32>;<low32>[;opt]" */
			{
				int len;
				
				/* Compute the size of the new string */
				len = strlen(g_pconf->diameter_identity);	/* This could be cached to avoid recomputing it each time */
				len += 2 * ( 1 /* ';' */ + 10 /* uint32_t max = 4294967295 = 10 digits */ );
				if (opt)
					len += strlen(opt);
				
				/* Allocate the space for the buffer */
				str = malloc(len + 1); /* +1 for the final '\0' in the worst case */
				if (!str) {
					log_error("Memory allocation failed: %s\n", strerror(errno));
					TRACE_DEBUG(INFO, "malloc failed");
					return ENOMEM;
				}
				mustfree = 1;
				memset(str, 0, len+1);
				
				/* Increment the value of the low32 */
				ret = pthread_mutex_lock(&g_sid_lock);
				if (ret != 0) {
					TRACE_DEBUG(INFO, "pthread_mutex_lock failed: %s", strerror(ret));
					goto end;
				}
				++g_sid_l;
				
				/* Now write the new value */
				if (opt)
					strsz = snprintf(str, len + 1, "%s;%u;%u;%s", g_pconf->diameter_identity, g_sid_h, g_sid_l, opt);
				else
					strsz = snprintf(str, len + 1, "%s;%u;%u", g_pconf->diameter_identity, g_sid_h, g_sid_l);
				
				ret = pthread_mutex_unlock(&g_sid_lock);
				if (ret != 0) {
					TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
					goto end;
				}
			}
			
			break;
			
		case SESSION_NEW_SEQUENCE: 	/* "<diameterId>;opt" */
			if (opt == NULL) {
				TRACE_DEBUG(INFO, "Invalid parameter");
				return EINVAL;
			}
			{
				int len;
				
				/* Compute the size of the new string */
				len = strlen(g_pconf->diameter_identity) + 1 + strlen(opt) + 1; /* +1 for the ';' and the final '\0' */
				
				/* Allocate the space for the buffer */
				str = malloc(len);
				if (!str) {
					log_error("Memory allocation failed: %s\n", strerror(errno));
					TRACE_DEBUG(INFO, "malloc failed");
					return ENOMEM;
				}
				mustfree = 1;
				memset(str, 0, len);
				
				strsz = snprintf(str, len, "%s;%s", g_pconf->diameter_identity, opt);
			}
			break;
			
		case SESSION_NEW_FULL:		/* "opt"  */
			if (opt == NULL) {
				TRACE_DEBUG(INFO, "Invalid parameter");
				return EINVAL;
			}
			str = opt;
			strsz = strlen(str);
			break;

		default:
			TRACE_DEBUG(INFO, "Invalid parameter");
			return EINVAL;
	}
	
	/* Now str points to the new session string. Create the new session */
	TRACE_DEBUG(FULL, "Creating session object for sid: '%s'", str);
	
	ret = sess_fromsid( str, strsz, session, &new );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "sess_fromsid failed: %s", strerror(ret));
		goto end;
	}
	
	/* Check that the object did not already exist before */
	if (new == 0) {
		TRACE_DEBUG(INFO, "The session created already existed => The opt parameter must have been bad");
		
		/* The string we received was not really a new unique value */
		ret = sess_unlink(session);
		if (ret != 0) {
			TRACE_DEBUG(INFO, "sess_unlink failed: %s", strerror(ret));
			goto end;
		}
		
		ret = EALREADY;
	}
	
end:
	if (mustfree)
		free(str);
	
	return ret;
}

/* Get the session-id string of a session */
int sess_getsid ( sess_id_t * session, char ** sid )
{
	TRACE_ENTRY( "%p %p", session, sid );
	
	/* Check the parameter */
	if ( (! VALIDATE_SI(session)) || (sid == NULL)) {
		TRACE_DEBUG(INFO, "Invalid parameter");
		return EINVAL;
	}
	
	/* Write the location and exit */
	*sid = _SI(session)->sid;
	return 0;
}

/* Increase refcount of a session */
int sess_link ( sess_id_t * session )
{
	int ret = 0;
	
	TRACE_ENTRY( "%p", session );
	
	/* Check the parameter */
	if (! VALIDATE_SI(session)) {
		TRACE_DEBUG(INFO, "Invalid parameter");
		return EINVAL;
	}
	
	/* Lock the list entry */
	ret = pthread_mutex_lock(H_LOCK(_SI(session)->hash));
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_lock failed: %s", strerror(ret));
		return ret;
	}
	
	/* Increment the refcount */
	_SI(session)->rc++;
	
	/* Unlock the list entry */
	ret = pthread_mutex_unlock(H_LOCK(_SI(session)->hash));
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
		return ret;
	}
	
	return 0;
}

/* Decrease refcount of a session, eventually free it. */
int sess_unlink ( sess_id_t ** session )
{
	int ret = 0;
	uint32_t hash;
	
	TRACE_ENTRY( "%p", session );
	
	/* Check the parameter */
	if ((session == NULL) || (! VALIDATE_SI(*session))) {
		TRACE_DEBUG(INFO, "Invalid parameter");
		return EINVAL;
	}
	
	/* Copy the hash value */
	hash = _SI(*session)->hash;
	
	/* Lock the list entry */
	ret = pthread_mutex_lock(H_LOCK(hash));
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_lock failed: %s", strerror(ret));
		return ret;
	}
	
	/* Decrement the refcount */
	_SI(*session)->rc--;
	
	/* If the count is now 0, we destroy the object */
	if (_SI(*session)->rc == 0) {
		_SI(*session)->eyec = 0xdead;
		
		/* Unlink the element from the chain */
		SLI_remove(*session); /* actually &(*session)->chain, which is the same */
		
		/* Update the count */
		H_COUNT(hash) -= 1;
		
		/* Now free the element */
		if (_SI(*session)->sid)
			free(_SI(*session)->sid);
		free(*session);
	}
	
	/* Unlock the list entry */
	ret = pthread_mutex_unlock(H_LOCK(hash));
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
		return ret;
	}
	
	/* Update the location for the calling function */
	*session = NULL;
	
	return 0;
}

/* Create a new client of the module. */
int sess_regext ( sess_reg_t ** client )
{
	int ret = 0;
	_sess_reg_t * new = NULL;
	
	TRACE_ENTRY( "%p", client );
	
	/* Check the parameter */
	if (client == NULL) {
		TRACE_DEBUG(INFO, "Invalid parameter");
		return EINVAL;
	}
	
	/* Create a new object */
	new = (_sess_reg_t *) malloc(sizeof(_sess_reg_t));
	if (new == NULL) {
		log_error("Memory allocation failed: %s\n", strerror(errno));
		TRACE_DEBUG(INFO, "malloc failed");
		return ENOMEM;
	}
	
	/* Initialize the object */
	memset(new, 0, sizeof(_sess_reg_t));
	SLI_init(&new->list);
	new->eyec = _SR_EYEC;
	ret = pthread_mutex_init(&new->lock, NULL);
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_init failed: %s", strerror(ret));
		free(new);
		return ret;
	}
	SLI_init(&new->sentinel);
	
	/* Save the new object in the list */
	ret = pthread_mutex_lock(&_str_lock);
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_lock failed: %s", strerror(ret));
		free(new);
		return ret;
	}
	
	SLI_addafter( &_srt_sentinel, &new->list);
	
	ret = pthread_mutex_unlock(&_str_lock);
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
		return ret;
	}
	
	/* Copy the reference to the calling process */
	*client = (sess_reg_t *)new;
	
	return 0;
}

/* Destroy a client of the module. */
int sess_deregext ( sess_reg_t * client )
{
	int ret = 0;
	
	TRACE_ENTRY( "%p", client );
	
	if (!VALIDATE_SR(client)) {
		TRACE_DEBUG(INFO, "Invalid parameter");
		return EINVAL;
	}
	
	/* Lock the list of clients */
	ret = pthread_mutex_lock( &_str_lock );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_lock failed: %s", strerror(ret));
		return ret;
	}
	
	/* Now destroy the client handler */
	ret = _destroy_sr(_SR(client));
	if (ret != 0) {
		TRACE_DEBUG(INFO, "_destroy_sr failed: %s", strerror(ret));
		(void)pthread_mutex_unlock( &_str_lock );
		return ret;
	}
	
	/* Unlock the list of clients */
	ret = pthread_mutex_unlock( &_str_lock );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
		return ret;
	}
	
	return 0;
}

/* Register data in a session. */
int sess_data_reg ( sess_id_t * session, sess_reg_t * client, void * data, void (*cleanup)(void *) )
{
	int ret = 0;
	_sd_data_t * new = NULL;
	 _sli_t *prev = NULL;
	
	TRACE_ENTRY( "%p %p %p %p", session, client, data, cleanup );
	
	/* Check parameters */
	if ( (!VALIDATE_SI(session)) || (!VALIDATE_SR(client)) || (data == NULL)) {
		TRACE_DEBUG(INFO, "Invalid parameter");
		return EINVAL;
	}
	
	/* Create a new object to hold the data */
	new = (_sd_data_t *) malloc(sizeof(_sd_data_t));
	if (new == NULL) {
		log_error("Memory allocation failed: %s\n", strerror(errno));
		TRACE_DEBUG(INFO, "malloc failed");
		return ENOMEM;
	}
	memset(new, 0, sizeof(_sd_data_t));
	SLI_init(&new->list);
	new->si = _SI(session);
	new->clncb = cleanup;
	new->data = data;
	
	/* Link the session */
	ret = sess_link((sess_id_t *)new->si);
	if (ret != 0) {
		TRACE_DEBUG(INFO, "sess_link failed: %s", strerror(ret));
		free(new);
		return ret;
	}
	
	/* Lock the client object */
	ret = pthread_mutex_lock( &_SR(client)->lock );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_lock failed: %s", strerror(ret));
		(void)sess_unlink((sess_id_t **)&new->si);
		free(new);
		return ret;
	}
	
	/* Find the place in the list */
	ret = find_sd(new->si, &_SR(client)->sentinel, &prev);
	if (ret == 0) {
		/* There is already data associated with this session and client */
		TRACE_DEBUG(INFO, "Data already associated with session/client.");
		(void)sess_unlink((sess_id_t **)&new->si);
		free(new);
		(void)pthread_mutex_unlock( &_SR(client)->lock );
		return EALREADY;
	}
	
	/* Now "prev" is the previous element where the new element must be put */
	SLI_addafter(prev, new);
	
	/* Unlock, we're done */
	ret = pthread_mutex_unlock( &_SR(client)->lock );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
		return ret;
	}
	
	return 0;
}

/* Retrieve registered data. */
int sess_data_get ( sess_id_t * session, sess_reg_t * client, void ** data )
{
	int ret = 0;
	 _sli_t *item = NULL;
	 
	TRACE_ENTRY( "%p %p %p", session, client, data );
	
	/* Check parameters */
	if ( (!VALIDATE_SI(session)) || (!VALIDATE_SR(client)) || (data == NULL)) {
		TRACE_DEBUG(INFO, "Invalid parameter");
		return EINVAL;
	}
	
	*data = NULL;
	
	/* Lock the client object */
	ret = pthread_mutex_lock( &_SR(client)->lock );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_lock failed: %s", strerror(ret));
		return ret;
	}
	
	/* Find the data */
	ret = find_sd(_SI(session), &_SR(client)->sentinel, &item);
	if (ret == 0) {
		*data = _SD(item)->data;
	}
	
	/* Unlock, we're done */
	ret = pthread_mutex_unlock( &_SR(client)->lock );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
		return ret;
	}
	
	return (*data == NULL) ? ENOENT : 0;
}

/* Remove registered data from a session. */
int sess_data_dereg ( sess_id_t * session, sess_reg_t * client, void ** data )
{
	int ret = 0;
	int noent = 0;
	 _sli_t *item = NULL;
	 
	TRACE_ENTRY( "%p %p %p", session, client, data );
	
	/* Check parameters */
	if ( (!VALIDATE_SI(session)) || (!VALIDATE_SR(client)) ) {
		TRACE_DEBUG(INFO, "Invalid parameter");
		return EINVAL;
	}
	
	if (data)
		*data = NULL;
	
	/* Lock the client object */
	ret = pthread_mutex_lock( &_SR(client)->lock );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_lock failed: %s", strerror(ret));
		return ret;
	}
	
	/* Find the data */
	ret = find_sd(_SI(session), &_SR(client)->sentinel, &item);
	if (ret == 0) {
		ret = _destroy_sr_data(_SD(item), data);
		if (ret != 0) {
			TRACE_DEBUG(INFO, "_destroy_sr_data failed: %s", strerror(ret));
			(void)pthread_mutex_unlock( &_SR(client)->lock );
			return ret;
		}
	} else {
		TRACE_DEBUG(INFO, "data not found.");
		noent = 1;
	}
	
	/* Unlock, we're done */
	ret = pthread_mutex_unlock( &_SR(client)->lock );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
		return ret;
	}
	
	return noent ? ENOENT : 0;
}

/* End of the module */
int sess_fini ( void )
{
	int ret = 0;
	int i;
	
	TRACE_ENTRY( "" );
	
	/* Lock the list of clients */
	ret = pthread_mutex_lock( &_str_lock );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_lock failed: %s", strerror(ret));
		return ret;
	}
	
	/* Destroy all the registered clients */
	while (!SLI_isempty( &_srt_sentinel )) {
		TRACE_DEBUG(FULL, "Destroying module %p...", _srt_sentinel.next);
		ret = _destroy_sr( _SR(_srt_sentinel.next) );
		if (ret != 0) {
			TRACE_DEBUG(INFO, "_destroy_sr failed: %s", strerror(ret));
			(void)pthread_mutex_unlock( &_str_lock );
			return ret;
		}
	}
	
	/* Unlock the list of clients */
	ret = pthread_mutex_unlock( &_str_lock );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
		return ret;
	}
	
	/* Destroy the lock */
	ret = pthread_mutex_destroy( &_str_lock );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_destroy failed: %s", strerror(ret));
		return ret;
	}
	
	/* Check that there is no more session registered */
	for (i = 0; i < sizeof(sess_hash) / sizeof(sess_hash[0]); i++) {
		if (!SLI_isempty(&sess_hash[i].sentinel)) {
			TRACE_DEBUG(INFO, "List of session (%d) is not empty. Maybe messages are still refering sessions? EBUSY", i);
			return EBUSY;
		}
		ret = pthread_mutex_destroy( &sess_hash[i].lock );
		if (ret != 0) {
			TRACE_DEBUG(INFO, "pthread_mutex_destroy failed: %s", strerror(ret));
			return ret;
		}
	}
	
	/* Done! */
	return 0;
}

/********************************************************************************************************/
/* debug functions */

static void _dump_si ( _sess_id_t * si, int indent )
{
	TRACE_DEBUG(FULL, "%*ssession @%p: h:%8x rc:%d ts:%d sid:'%s'", indent, "", si, si->hash, si->rc, si->ts.tv_sec, si->sid);
}

static void _dump_hash (void)
{
	int i;
	
	TRACE_DEBUG(FULL, " ======= Dumping the session hash table ======= ");
	for (i=0; i < sizeof(sess_hash) / sizeof(sess_hash[0]); i++) {
		_sli_t * item;
		TRACE_DEBUG(FULL, " l:%02d cnt:%d", i, sess_hash[i].count);
		for (item = &sess_hash[i].sentinel; item->next != &sess_hash[i].sentinel; item = item->next)
			_dump_si(_SI(item->next), 3);
	}
	TRACE_DEBUG(FULL, " =======/end of session hash table dump/======= ");
}	

static void _dump_sd ( _sd_data_t * sd, int indent )
{
	TRACE_DEBUG(FULL, "%*sdata @%p: data:%p clncb:%p sid:'%s'", indent, "", sd, sd->data, sd->clncb, sd->si->sid);
}

static void _dump_regs (void)
{
	_sli_t * client;
	_sli_t * data;
	
	TRACE_DEBUG(FULL, " ======= Dumping the registered clients list ======= ");
	for (client = &_srt_sentinel; client->next != &_srt_sentinel; client = client->next) {
		TRACE_DEBUG(FULL, "client @%p:", client->next);
		for (data = &_SR(client->next)->sentinel; data->next != &_SR(client->next)->sentinel; data = data->next) {
			_dump_sd(_SD(data->next), 3);
		}
	}
	TRACE_DEBUG(FULL, " =======/end of registered clients list dump/======= ");
}	

void sess_dump(void) 
{
	_dump_hash();
	_dump_regs();
}

